/*
 * myfunc
 *
 * Sample to demonstrate C functions
 *
 * Joe Conway <mail@joeconway.com>
 *
 * Copyright (c) 2004, Joseph E. Conway
 * ALL RIGHTS RESERVED
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without a written agreement
 * is hereby granted, provided that the above copyright notice and this
 * paragraph and the following two paragraphs appear in all copies.
 * 
 * IN NO EVENT SHALL THE AUTHORS OR DISTRIBUTORS BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING
 * LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS
 * DOCUMENTATION, EVEN IF THE AUTHOR OR DISTRIBUTORS HAVE BEEN ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE AUTHORS AND DISTRIBUTORS SPECIFICALLY DISCLAIM ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE AUTHOR AND DISTRIBUTORS HAS NO OBLIGATIONS TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 *
 */
#include "postgres.h"

#include "fmgr.h"
#include "funcapi.h"
#include "executor/spi.h"
#include "lib/stringinfo.h"
#include "miscadmin.h"
#include "utils/array.h"
#include "utils/builtins.h"
#include "utils/hsearch.h"
#include "utils/lsyscache.h"

/*
 * exported functions
 */
extern Datum extract_field_as_text(PG_FUNCTION_ARGS);
extern Datum create_array(PG_FUNCTION_ARGS);
extern Datum build_tuples_from_textarray(PG_FUNCTION_ARGS);
extern Datum build_tuple_from_textarray(PG_FUNCTION_ARGS);
extern Datum myfunc_errcontext(PG_FUNCTION_ARGS);
extern Datum build_tuples_from_textarray_funcapi(PG_FUNCTION_ARGS);
extern Datum myfunc_setvar(PG_FUNCTION_ARGS);
extern Datum myfunc_getvar(PG_FUNCTION_ARGS);
extern Datum myfunc_rmvar(PG_FUNCTION_ARGS);


/*
 * internal
 */
static void myfunc_error_callback(void *arg);
static char *getStrByName(const char *name);
static HTAB *createStrHash(void);
static void createNewStr(const char *name, char *str);
static void deleteStr(const char *name);

static HTAB *strHash = NULL;
typedef struct strHashEnt
{
	char		name[NAMEDATALEN];
	char	   *str;
}	strHashEnt;


/*
 * definitions
 */
#define NUMSTR 16
#define PG_CSTR_GET_TEXT(cstrp) DatumGetTextP(DirectFunctionCall1(textin, CStringGetDatum(cstrp)))
#define PG_TEXT_GET_CSTR(textp) DatumGetCString(DirectFunctionCall1(textout, PointerGetDatum(textp)))


/*
 * Example to demonstrate a composite argument. Accepts the composite
 * argument and a field name, returns the field value as a scalar text.
 *
 * NOTE: this version works with Postgres 7.4 and earlier. Postgres
 * 7.5 support would require a few changes.
 */
PG_FUNCTION_INFO_V1(extract_field_as_text);
Datum
extract_field_as_text(PG_FUNCTION_ARGS)
{
	TupleTableSlot *slot;
	char		   *desired_col_name;
	HeapTuple		tuple;
	TupleDesc		tupdesc;
	int				nc;
	int				j = 0;

	/* Get the tuple argument and initialize a few things */
	slot = (TupleTableSlot *) PG_GETARG_POINTER(0);
	tuple = slot->val;
	tupdesc = slot->ttc_tupleDescriptor;
	nc = tupdesc->natts;

	/* Get the field name to extract */
	desired_col_name = PG_TEXT_GET_CSTR(PG_GETARG_TEXT_P(1));

	/*
	 * Loop by columns
	 */
	for (j = 0; j < nc; j++)		
	{
		char	   *this_col_name;

		/* ignore dropped attributes */
		if (tupdesc->attrs[j]->attisdropped)
			continue;

		/* get column name */
		this_col_name = SPI_fname(tupdesc, j + 1);

		/* is it the one we're looking for? */
		if (strcmp(this_col_name, desired_col_name) == 0)
		{
				char	   *value;

				value = SPI_getvalue(tuple, tupdesc, j + 1);
				if (value)
			/*		PG_RETURN_TEXT_P(PG_CSTR_GET_TEXT(value));		*/
					PG_RETURN_DATUM(DirectFunctionCall1(textin,
														CStringGetDatum(value)));
				else
					PG_RETURN_NULL();
		}
	}

	/* column was not found -- complain */
	ereport(ERROR,
			(errcode(ERRCODE_INVALID_PARAMETER_VALUE),
			 errmsg("column not found: %s", desired_col_name)));

	/* never reached -- keep the compiler happy */
	PG_RETURN_NULL();
}

/*
 * Example to demonstrate forming and returning an array
 */
PG_FUNCTION_INFO_V1(create_array);
Datum
create_array(PG_FUNCTION_ARGS)
{
	ArrayType  *v;
	Datum		dvalues[1];
	int16		typlen;
	bool		typbyval;
	char		typalign;
	int			dims[MAXDIM];
	int			lbs[MAXDIM];
	int			i;
	Oid			element_type;
	Oid			array_type;
	int			ndims;

	if (PG_NARGS() == 2)
	{
		/* get requested number of dimensions for the array */
		ndims = PG_GETARG_INT32(0);

		/* see what type of element we were passed */
		element_type = get_fn_expr_argtype(fcinfo->flinfo, 1);

		/* get the actual array element */
		dvalues[0] = PG_GETARG_DATUM(1);
	}
	else
	{
		/* if not specified, assume 1D array */
		ndims = 1;

		/* see what type of element we were passed */
		element_type = get_fn_expr_argtype(fcinfo->flinfo, 0);

		/* get the actual array element */
		dvalues[0] = PG_GETARG_DATUM(0);
	}

	/* check that an array of this element type is actually valid */
	array_type = get_array_type(element_type);
	if (array_type == InvalidOid)
		ereport(ERROR,
				(errcode(ERRCODE_INVALID_PARAMETER_VALUE),
				 errmsg("invalid array element type OID: %u", element_type)));

	/* initialize number of elements and lower bound for each dimension */
	for (i = 0; i < ndims; i++)
	{
		dims[i] = 1;
		lbs[i] = 1;
	}

	/* get info about element type needed to construct the array */
	get_typlenbyvalalign(element_type, &typlen, &typbyval, &typalign);

	/* construct it */
	v = construct_md_array(dvalues, ndims, dims, lbs, element_type,
									   typlen, typbyval, typalign);

	/* return it */
	PG_RETURN_ARRAYTYPE_P(v);
}


/*
 * Example to demonstrate a composite return type. Accepts an array
 * of text, returns the array elements formed into a tuple according
 * to the type specified at runtime.
 *
 * If the input array is 1D, return a single tuple. If 2D, return
 * the elements of the inner dimension as multiple rows. The
 * cardinality of the outer (first) dimension determines the number
 * of rows to be returned.
 */
PG_FUNCTION_INFO_V1(build_tuples_from_textarray);
Datum
build_tuples_from_textarray(PG_FUNCTION_ARGS)
{
	ArrayType	   *v = PG_GETARG_ARRAYTYPE_P(0);
	Oid				element_type = ARR_ELEMTYPE(v);
	int				ndims = ARR_NDIM(v);
	int			   *dim_counts = ARR_DIMS(v);
	int			   *dim_lower_bounds = ARR_LBOUND(v);
	int				ncols = 0;
	int				nrows = 0;
	int				indx[MAXDIM];
	int16			typlen;
	bool			typbyval;
	char			typalign;
	ReturnSetInfo  *rsinfo = (ReturnSetInfo *) fcinfo->resultinfo;
	AttInMetadata  *attinmeta;
	TupleDesc		tupdesc;
	Tuplestorestate *tupstore = NULL;
	HeapTuple		tuple;
	MemoryContext	per_query_ctx;
	MemoryContext	oldcontext;
	Datum			dvalue;
	char		  **values;
	int				rsinfo_ncols;
	int				i, j;

	/* only allow 1D or 2D input array */
	if (ndims == 1)
	{
		nrows = 1;
		ncols = dim_counts[0];
	}
	else if (ndims == 2)
	{
		nrows = dim_counts[0];
		ncols = dim_counts[1];
	}
	else
		ereport(ERROR,
				(errcode(ERRCODE_INVALID_PARAMETER_VALUE),
				 errmsg("invalid input array"),
				 errdetail("Input array must be 1 or 2 dimensions.")));
		
	/* check to see if caller supports us returning a tuplestore */
	if (!rsinfo || !(rsinfo->allowedModes & SFRM_Materialize))
		ereport(ERROR,
				(errcode(ERRCODE_SYNTAX_ERROR),
				 errmsg("materialize mode required, but it is not " \
						"allowed in this context")));

	/* get info about element type needed to construct the array */
	get_typlenbyvalalign(element_type, &typlen, &typbyval, &typalign);

	/* get the requested return tuple description */
	tupdesc = rsinfo->expectedDesc;
	rsinfo_ncols = tupdesc->natts;

	/*
	 * The requested tuple description better match up with the array
	 * we were given.
	 */
	if (rsinfo_ncols != ncols)
		ereport(ERROR,
				(errcode(ERRCODE_INVALID_PARAMETER_VALUE),
				 errmsg("invalid input array"),
				 errdetail("Number of elements in array inner dimension must "
						   "match number of query specified columns.")));

	/* OK, use it */
	attinmeta = TupleDescGetAttInMetadata(tupdesc);

	/* Now go to work */
	rsinfo->returnMode = SFRM_Materialize;

	per_query_ctx = fcinfo->flinfo->fn_mcxt;
	oldcontext = MemoryContextSwitchTo(per_query_ctx);

	/* initialize our tuplestore */
	tupstore = tuplestore_begin_heap(true, false, SortMem);

	values = (char **) palloc(ncols * sizeof(char *));

	for (i = 0; i < nrows; i++)
	{
		for (j = 0; j < ncols; j++)
		{
			bool	isnull;

			/* array value of this position */
			if (ndims == 1)
				indx[0] = j + dim_lower_bounds[0];
			else
			{
				indx[0] = i + dim_lower_bounds[0];
				indx[1] = j + dim_lower_bounds[1];
			}
			dvalue = array_ref(v, ndims, indx, -1, typlen,
								 typbyval, typalign, &isnull);

			if (!isnull)
				values[j] = DatumGetCString(DirectFunctionCall1(textout,
																dvalue));
			else
				values[j] = NULL;
		}
		/* construct the tuple */
		tuple = BuildTupleFromCStrings(attinmeta, values);

		/* now store it */
		tuplestore_puttuple(tupstore, tuple);
	}

	tuplestore_donestoring(tupstore);
	rsinfo->setResult = tupstore;

	/*
	 * SFRM_Materialize mode expects us to return a NULL Datum. The actual
	 * tuples are in our tuplestore and passed back through
	 * rsinfo->setResult. rsinfo->setDesc is set to the tuple description
	 * that we actually used to build our tuples with, so the caller can
	 * verify we did what it was expecting.
	 */
	rsinfo->setDesc = tupdesc;
	MemoryContextSwitchTo(oldcontext);

	return (Datum) 0;
}

/*
 * Similar to the above, but only deals with 1D array, and most error
 * handling is stripped out for the sake of brevity.
 */
PG_FUNCTION_INFO_V1(build_tuple_from_textarray);
Datum
build_tuple_from_textarray(PG_FUNCTION_ARGS)
{
	ArrayType	   *v = PG_GETARG_ARRAYTYPE_P(0);
	Oid				element_type = ARR_ELEMTYPE(v);
	int				ndims = ARR_NDIM(v);
	int			   *dim_counts = ARR_DIMS(v);
	int			   *dim_lower_bounds = ARR_LBOUND(v);
	int				ncols = 0;
	int				nrows = 0;
	int				indx[MAXDIM];
	int16			typlen;
	bool			typbyval;
	char			typalign;
	ReturnSetInfo  *rsinfo = (ReturnSetInfo *) fcinfo->resultinfo;
	AttInMetadata  *attinmeta;
	TupleDesc		tupdesc;
	Tuplestorestate *tupstore = NULL;
	HeapTuple		tuple;
	MemoryContext	per_query_ctx;
	MemoryContext	oldcontext;
	Datum			dvalue;
	char		  **values;
	int				j;

	if (ndims != 1)
		elog(ERROR, "input array must be 1D");
	if (!rsinfo || !(rsinfo->allowedModes & SFRM_Materialize))
		elog(ERROR, "materialize mode required, but it "
						"is not allowed in this context");

	nrows = 1;
	ncols = dim_counts[0];

	get_typlenbyvalalign(element_type, &typlen, &typbyval, &typalign);

	tupdesc = rsinfo->expectedDesc;
	attinmeta = TupleDescGetAttInMetadata(tupdesc);

	rsinfo->returnMode = SFRM_Materialize;

	per_query_ctx = fcinfo->flinfo->fn_mcxt;
	oldcontext = MemoryContextSwitchTo(per_query_ctx);
	tupstore = tuplestore_begin_heap(true, false, SortMem);

	values = (char **) palloc(ncols * sizeof(char *));

	for (j = 0; j < ncols; j++)
	{
		bool	isnull;

		indx[0] = j + dim_lower_bounds[0];
		dvalue = array_ref(v, ndims, indx, -1, typlen,
							typbyval, typalign, &isnull);

		values[j] = DatumGetCString(DirectFunctionCall1(textout, dvalue));
	}

	tuple = BuildTupleFromCStrings(attinmeta, values);
	tuplestore_puttuple(tupstore, tuple);
	tuplestore_donestoring(tupstore);

	rsinfo->setResult = tupstore;
	rsinfo->setDesc = tupdesc;

	MemoryContextSwitchTo(oldcontext);

	return (Datum) 0;
}


/*
 * Example to demonstrate a composite return type. Accepts an array
 * of text, returns the array elements formed into a tuple according
 * to the type specified at runtime.
 *
 * If the input array is 1D, return a single tuple. If 2D, return
 * the elements of the inner dimension as multiple rows. The
 * cardinality of the outer (first) dimension determines the number
 * of rows to be returned.
 *
 * Same as build_tuples_from_textarray, except uses funcapi
 * (one row at a time) instead of a tuplestore.
 */

typedef struct
{
	ArrayType		   *v;
	int16				typlen;
	bool				typbyval;
	char				typalign;
}	myfunc_fctx;

PG_FUNCTION_INFO_V1(build_tuples_from_textarray_funcapi);
Datum
build_tuples_from_textarray_funcapi(PG_FUNCTION_ARGS)
{
	FuncCallContext	   *funcctx;
	myfunc_fctx		   *fctx;
	ArrayType		   *v;
	int					ndims;
	int				   *dim_counts;
	int					ncols = 0;
	int16				typlen;
	bool				typbyval;
	char				typalign;
	ReturnSetInfo	   *rsinfo = (ReturnSetInfo *) fcinfo->resultinfo;
	AttInMetadata	   *attinmeta;
	TupleDesc			tupdesc;
	HeapTuple			tuple;
	TupleTableSlot	   *slot;

	if (SRF_IS_FIRSTCALL())
	{
		int				rsinfo_ncols;
		int				nrows = 0;
		Oid				element_type;
		MemoryContext	oldcontext;

		funcctx = SRF_FIRSTCALL_INIT();

		/*
		 * switch to memory context appropriate for multiple function
		 * calls
		 */
		oldcontext = MemoryContextSwitchTo(funcctx->multi_call_memory_ctx);

		/* get copy of the array and info about it */
		v = PG_GETARG_ARRAYTYPE_P_COPY(0);
		element_type = ARR_ELEMTYPE(v);
		ndims = ARR_NDIM(v);
		dim_counts = ARR_DIMS(v);
		
		/* only allow 1D or 2D input array */
		if (ndims == 1)
		{
			nrows = 1;
			ncols = dim_counts[0];
		}
		else if (ndims == 2)
		{
			nrows = dim_counts[0];
			ncols = dim_counts[1];
		}
		else
			ereport(ERROR,
					(errcode(ERRCODE_INVALID_PARAMETER_VALUE),
					 errmsg("invalid input array"),
					 errdetail("Input array must be 1 or 2 dimensions.")));
			
		/* get info about element type needed to construct the array */
		get_typlenbyvalalign(element_type, &typlen, &typbyval, &typalign);
	
		/* get the requested return tuple description */
		tupdesc = rsinfo->expectedDesc;
		rsinfo_ncols = tupdesc->natts;
	
		/*
		 * The requested tuple description better match up with the array
		 * we were given.
		 */
		if (rsinfo_ncols != ncols)
			ereport(ERROR,
					(errcode(ERRCODE_INVALID_PARAMETER_VALUE),
					 errmsg("invalid input array"),
					 errdetail("number of elements in array inner dimension must "
							   "match number of query specified columns.")));
	
		/* allocate a slot for a tuple with this tupdesc */
		slot = TupleDescGetSlot(tupdesc);
	
		/*
		 * Generate attribute metadata needed later to produce tuples from
		 * raw C strings
		 */
		attinmeta = TupleDescGetAttInMetadata(tupdesc);

		/* allocate memory for user context */
		fctx = (myfunc_fctx *) palloc(sizeof(myfunc_fctx));
	
		/*
		 * Save data for use across calls
		 */
		fctx->typlen = typlen;
		fctx->typbyval = typbyval;
		fctx->typalign = typalign;
		fctx->v = v;

		funcctx->user_fctx = (void *) fctx;
		funcctx->attinmeta = attinmeta;
		funcctx->slot = slot;
		funcctx->max_calls = nrows;

		MemoryContextSwitchTo(oldcontext);
	}

	funcctx = SRF_PERCALL_SETUP();

	/* get saved multicall data */
	slot = funcctx->slot;
	attinmeta = funcctx->attinmeta;
	tupdesc = attinmeta->tupdesc;
	fctx = (myfunc_fctx *) funcctx->user_fctx;

	typlen = fctx->typlen;
	typbyval = fctx->typbyval;
	typalign = fctx->typalign;
	v = fctx->v;
	ndims = ARR_NDIM(v);
	dim_counts = ARR_DIMS(v);

	if (ndims == 1)
		ncols = dim_counts[0];
	else if (ndims == 2)
		ncols = dim_counts[1];

	rsinfo->returnMode = SFRM_ValuePerCall;
	rsinfo->setDesc = tupdesc;

	if (funcctx->call_cntr < funcctx->max_calls)
	{
		Datum		dvalue;
		Datum		result;
		int		   *dim_lower_bounds = ARR_LBOUND(v);
		int			indx[MAXDIM];
		int			j;
		int			i = funcctx->call_cntr;
		char	  **values = (char **) palloc(ncols * sizeof(char *));

		for (j = 0; j < ncols; j++)
		{
			bool	isnull;

			/* array value of this position */
			if (ndims == 1)
				indx[0] = j + dim_lower_bounds[0];
			else
			{
				indx[0] = i + dim_lower_bounds[0];
				indx[1] = j + dim_lower_bounds[1];
			}
			dvalue = array_ref(v, ndims, indx, -1, typlen,
								 typbyval, typalign, &isnull);

			if (!isnull)
				values[j] = DatumGetCString(DirectFunctionCall1(textout,
																dvalue));
			else
				values[j] = NULL;
		}
		/* construct the tuple */
		tuple = BuildTupleFromCStrings(attinmeta, values);

		/* make the tuple into a datum */
		result = TupleGetDatum(slot, tuple);

		SRF_RETURN_NEXT(funcctx, result);
	}
	else
		SRF_RETURN_DONE(funcctx);
}


/*
 * Demonstrate use of error context callbacks
 */
PG_FUNCTION_INFO_V1(myfunc_errcontext);
Datum
myfunc_errcontext(PG_FUNCTION_ARGS)
{
	char				   *msg = PG_TEXT_GET_CSTR(PG_GETARG_TEXT_P(0));
	ErrorContextCallback	errcontext;

	errcontext.callback = myfunc_error_callback;
	errcontext.arg = (void *) "myfunc_errcontext";
	errcontext.previous = error_context_stack;
	error_context_stack = &errcontext;

	/* throw ERROR every time */
	elog(ERROR, "%s", msg);

	/*
	 * Never reached, but this is how to cleanup your callback when no
	 * longer needed. Also the return is there to satisfy the compiler.
	 */
	error_context_stack = errcontext.previous;
	PG_RETURN_VOID();
}


/*
 * Create a named persistent string
 */
PG_FUNCTION_INFO_V1(myfunc_setvar);
Datum
myfunc_setvar(PG_FUNCTION_ARGS)
{
	char	   *name = NULL;
	char	   *str = NULL;

	name = PG_TEXT_GET_CSTR(PG_GETARG_TEXT_P(0));
	str = PG_TEXT_GET_CSTR(PG_GETARG_TEXT_P(1));

	createNewStr(name, str);

	PG_RETURN_TEXT_P(PG_CSTR_GET_TEXT("OK"));
}

/*
 * Retrieve a named persistent string
 */
PG_FUNCTION_INFO_V1(myfunc_getvar);
Datum
myfunc_getvar(PG_FUNCTION_ARGS)
{
	char	   *name = NULL;
	char	   *str;

	name = PG_TEXT_GET_CSTR(PG_GETARG_TEXT_P(0));

	str = getStrByName(name);
	if (str)
		PG_RETURN_TEXT_P(PG_CSTR_GET_TEXT(str));
	else
		PG_RETURN_NULL();
}

/*
 * Delete a named persistent string
 */
PG_FUNCTION_INFO_V1(myfunc_rmvar);
Datum
myfunc_rmvar(PG_FUNCTION_ARGS)
{
	char	   *name = NULL;

	name = PG_TEXT_GET_CSTR(PG_GETARG_TEXT_P(0));

	deleteStr(name);

	PG_RETURN_TEXT_P(PG_CSTR_GET_TEXT("OK"));
}



/*****************************************************************
 * local functions
 */
static void
myfunc_error_callback(void *arg)
{
	if (arg)
		errcontext("In myfunc function %s", (char *) arg);
}

static char *
getStrByName(const char *name)
{
	strHashEnt *hentry;
	char		key[NAMEDATALEN];

	if (!strHash)
		strHash = createStrHash();

	MemSet(key, 0, NAMEDATALEN);
	snprintf(key, NAMEDATALEN - 1, "%s", name);
	hentry = (strHashEnt *) hash_search(strHash, key, HASH_FIND, NULL);

	if (hentry)
		return (hentry->str);

	return (NULL);
}

static HTAB *
createStrHash(void)
{
	HASHCTL		ctl;
	HTAB	   *ptr;

	ctl.keysize = NAMEDATALEN;
	ctl.entrysize = sizeof(strHashEnt);

	ptr = hash_create("str hash", NUMSTR, &ctl, HASH_ELEM);

	if (!ptr)
		ereport(ERROR,
				(errcode(ERRCODE_OUT_OF_MEMORY),
				 errmsg("out of memory")));

	return (ptr);
}

static void
createNewStr(const char *name, char *str)
{
	strHashEnt *hentry;
	bool		found;
	char		key[NAMEDATALEN];

	if (!strHash)
		strHash = createStrHash();

	MemSet(key, 0, NAMEDATALEN);
	snprintf(key, NAMEDATALEN - 1, "%s", name);
	hentry = (strHashEnt *) hash_search(strHash, key, HASH_ENTER, &found);

	if (!hentry)
		ereport(ERROR,
				(errcode(ERRCODE_OUT_OF_MEMORY),
				 errmsg("out of memory")));

	if (found)
		ereport(ERROR,
				(errcode(ERRCODE_DUPLICATE_OBJECT),
				 errmsg("duplicate string name")));

	hentry->str = MemoryContextStrdup(TopMemoryContext, str);
	strncpy(hentry->name, name, NAMEDATALEN - 1);
}

static void
deleteStr(const char *name)
{
	strHashEnt *hentry;
	bool		found;
	char		key[NAMEDATALEN];

	if (!strHash)
		strHash = createStrHash();

	MemSet(key, 0, NAMEDATALEN);
	snprintf(key, NAMEDATALEN - 1, "%s", name);

	hentry = (strHashEnt *) hash_search(strHash, key, HASH_REMOVE, &found);

	if (!hentry)
		ereport(ERROR,
				(errcode(ERRCODE_UNDEFINED_OBJECT),
				 errmsg("undefined string name")));

	if (hentry->str)
		pfree(hentry->str);
}
